@@ -60,15 +60,11 @@
 #define GSD_MEDIA_KEYS_DBUS_PATH GSD_DBUS_PATH "/MediaKeys"
 #define GSD_MEDIA_KEYS_DBUS_NAME GSD_DBUS_NAME ".MediaKeys"
 
+#define TOUCHPAD_ENABLED_KEY "/desktop/gnome/peripherals/touchpad/touchpad_enabled"
+
 #define VOLUME_STEP 6           /* percents for one volume button press */
 #define MAX_VOLUME 65536.0
 
-#if defined(__OpenBSD__)
-# define EJECT_COMMAND "eject -t /dev/rcd0c"
-#else
-# define EJECT_COMMAND "eject -T"
-#endif
-
 #define GSD_MEDIA_KEYS_MANAGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GSD_TYPE_MEDIA_KEYS_MANAGER, GsdMediaKeysManagerPrivate))
 
 typedef struct {
@@ -85,6 +81,7 @@ struct GsdMediaKeysManagerPrivate
 #endif /* HAVE_PULSE */
         GtkWidget       *dialog;
         GConfClient     *conf_client;
+        GVolumeMonitor  *volume_monitor;
 
         /* Multihead stuff */
         GdkScreen       *current_screen;
@@ -245,18 +242,6 @@ execute (GsdMediaKeysManager *manager,
         g_free (exec);
 }
 
-static void
-do_sleep_action (char *cmd1,
-                 char *cmd2)
-{
-        if (g_spawn_command_line_async (cmd1, NULL) == FALSE) {
-                if (g_spawn_command_line_async (cmd2, NULL) == FALSE) {
-                        acme_error (_("Couldn't put the machine to sleep.\n"
-                                        "Verify that the machine is correctly configured."));
-                }
-        }
-}
-
 static void
 dialog_init (GsdMediaKeysManager *manager)
 {
@@ -584,34 +569,93 @@ do_exit_action (GsdMediaKeysManager *manager)
         execute (manager, "gnome-session-save --shutdown-dialog", FALSE, FALSE);
 }
 
+static void
+do_eject_action_cb (GDrive              *drive,
+                    GAsyncResult        *res,
+                    GsdMediaKeysManager *manager)
+{
+        g_drive_eject_with_operation_finish (drive, res, NULL);
+}
+
+#define NO_SCORE 0
+#define SCORE_CAN_EJECT 50
+#define SCORE_HAS_MEDIA 100
 static void
 do_eject_action (GsdMediaKeysManager *manager)
 {
-        char *command;
+        GList *drives, *l;
+        GDrive *fav_drive;
+        guint score;
 
+        /* Find the best drive to eject */
+        fav_drive = NULL;
+        score = NO_SCORE;
+        drives = g_volume_monitor_get_connected_drives (manager->priv->volume_monitor);
+        for (l = drives; l != NULL; l = l->next) {
+                GDrive *drive = l->data;
+
+                if (g_drive_can_eject (drive) == FALSE)
+                        continue;
+                if (g_drive_is_media_removable (drive) == FALSE)
+                        continue;
+                if (score < SCORE_CAN_EJECT) {
+                        fav_drive = drive;
+                        score = SCORE_CAN_EJECT;
+                }
+                if (g_drive_has_media (drive) == FALSE)
+                        continue;
+                if (score < SCORE_HAS_MEDIA) {
+                        fav_drive = drive;
+                        score = SCORE_HAS_MEDIA;
+                        break;
+                }
+        }
+
+        /* Show the dialogue */
         dialog_init (manager);
         gsd_media_keys_window_set_action_custom (GSD_MEDIA_KEYS_WINDOW (manager->priv->dialog),
                                                  "media-eject",
                                                  FALSE);
         dialog_show (manager);
 
-        command = gconf_client_get_string (manager->priv->conf_client,
-                                           GCONF_MISC_DIR "/eject_command",
-                                           NULL);
-        if ((command != NULL) && (strcmp (command, "") != 0)) {
-                execute (manager, command, FALSE, FALSE);
-        } else {
-                execute (manager, EJECT_COMMAND, FALSE, FALSE);
-        }
+        /* Clean up the drive selection and exit if no suitable
+         * drives are found */
+        if (fav_drive != NULL)
+                fav_drive = g_object_ref (fav_drive);
 
-        g_free (command);
+        g_list_foreach (drives, (GFunc) g_object_unref, NULL);
+        if (fav_drive == NULL)
+                return;
+
+        /* Eject! */
+        g_drive_eject_with_operation (fav_drive, G_MOUNT_UNMOUNT_FORCE,
+                                      NULL, NULL,
+                                      (GAsyncReadyCallback) do_eject_action_cb,
+                                      manager);
+        g_object_unref (fav_drive);
+}
+
+static void
+do_touchpad_action (GsdMediaKeysManager *manager)
+{
+        GConfClient *client = manager->priv->conf_client;
+        gboolean state = gconf_client_get_bool (client, TOUCHPAD_ENABLED_KEY, NULL);
+
+        dialog_init (manager);
+        gsd_media_keys_window_set_action_custom (GSD_MEDIA_KEYS_WINDOW (manager->priv->dialog),
+                                                 (!state) ? "touchpad-enabled" : "touchpad-disabled",
+                                                 FALSE);
+        dialog_show (manager);
+
+        gconf_client_set_bool (client, TOUCHPAD_ENABLED_KEY, !state, NULL);
 }
 
 #ifdef HAVE_PULSE
 static void
 update_dialog (GsdMediaKeysManager *manager,
                guint vol,
-               gboolean muted)
+               gboolean muted,
+               gboolean sound_changed)
 {
         vol = (int) (100 * (double) vol / PA_VOLUME_NORM);
         vol = CLAMP (vol, 0, 100);
@@ -624,11 +668,12 @@ update_dialog (GsdMediaKeysManager *manager,
                                           GSD_MEDIA_KEYS_WINDOW_ACTION_VOLUME);
         dialog_show (manager);
 
-        ca_gtk_play_for_widget (manager->priv->dialog, 0,
-                                CA_PROP_EVENT_ID, "audio-volume-change",
-                                CA_PROP_EVENT_DESCRIPTION, "volume changed through key press",
-                                CA_PROP_APPLICATION_ID, "org.gnome.VolumeControl",
-                                NULL);
+        if (sound_changed != FALSE && muted == FALSE)
+                ca_gtk_play_for_widget (manager->priv->dialog, 0,
+                                        CA_PROP_EVENT_ID, "audio-volume-change",
+                                        CA_PROP_EVENT_DESCRIPTION, "volume changed through key press",
+                                        CA_PROP_APPLICATION_ID, "org.gnome.VolumeControl",
+                                        NULL);
 }
 
 static void
@@ -638,6 +683,7 @@ do_sound_action (GsdMediaKeysManager *manager,
         gboolean muted;
         guint vol, norm_vol_step;
         int vol_step;
+        gboolean sound_changed;
 
         if (manager->priv->stream == NULL)
                 return;
@@ -654,23 +700,29 @@ do_sound_action (GsdMediaKeysManager *manager,
         /* FIXME: this is racy */
         vol = gvc_mixer_stream_get_volume (manager->priv->stream);
         muted = gvc_mixer_stream_get_is_muted (manager->priv->stream);
+        sound_changed = FALSE;
 
         switch (type) {
         case MUTE_KEY:
                 muted = !muted;
                 gvc_mixer_stream_change_is_muted (manager->priv->stream, muted);
+                sound_changed = TRUE;
                 break;
         case VOLUME_DOWN_KEY:
                 if (!muted && (vol <= norm_vol_step)) {
                         muted = !muted;
                         vol = 0;
                         gvc_mixer_stream_change_is_muted (manager->priv->stream, muted);
-                        if (gvc_mixer_stream_set_volume (manager->priv->stream, vol) != FALSE)
+                        if (gvc_mixer_stream_set_volume (manager->priv->stream, vol) != FALSE) {
                                 gvc_mixer_stream_push_volume (manager->priv->stream);
+                                sound_changed = TRUE;
+                        }
                 } else if (!muted) {
                         vol = vol - norm_vol_step;
-                        if (gvc_mixer_stream_set_volume (manager->priv->stream, vol) != FALSE)
+                        if (gvc_mixer_stream_set_volume (manager->priv->stream, vol) != FALSE) {
                                 gvc_mixer_stream_push_volume (manager->priv->stream);
+                                sound_changed = TRUE;
+                        }
                 }
                 break;
         case VOLUME_UP_KEY:
@@ -679,10 +731,13 @@ do_sound_action (GsdMediaKeysManager *manager,
                         if (vol == 0) {
                                vol = vol + norm_vol_step;
                                gvc_mixer_stream_change_is_muted (manager->priv->stream, muted);
-                               if (gvc_mixer_stream_set_volume (manager->priv->stream, vol) != FALSE)
+                               if (gvc_mixer_stream_set_volume (manager->priv->stream, vol) != FALSE) {
                                         gvc_mixer_stream_push_volume (manager->priv->stream);
+                                        sound_changed = TRUE;
+                               }
                         } else {
                                 gvc_mixer_stream_change_is_muted (manager->priv->stream, muted);
+                                sound_changed = TRUE;
                         }
                 } else {
                         if (vol < MAX_VOLUME) {
@@ -691,14 +746,16 @@ do_sound_action (GsdMediaKeysManager *manager,
                                 } else {
                                         vol = vol + norm_vol_step;
                                 }
-                                if (gvc_mixer_stream_set_volume (manager->priv->stream, vol) != FALSE)
+                                if (gvc_mixer_stream_set_volume (manager->priv->stream, vol) != FALSE) {
                                         gvc_mixer_stream_push_volume (manager->priv->stream);
+                                        sound_changed = TRUE;
+                                }
                         }
                 }
                 break;
         }
 
-        update_dialog (manager, vol, muted);
+        update_dialog (manager, vol, muted, sound_changed);
 }
 
 static void
@@ -856,6 +913,9 @@ do_action (GsdMediaKeysManager *manager,
         char *path;
 
         switch (type) {
+        case TOUCHPAD_KEY:
+                do_touchpad_action (manager);
+                break;
         case MUTE_KEY:
         case VOLUME_DOWN_KEY:
         case VOLUME_UP_KEY:
@@ -1004,6 +1064,7 @@ start_media_keys_idle_cb (GsdMediaKeysManager *manager)
 
         g_debug ("Starting media_keys manager");
         gnome_settings_profile_start (NULL);
+        manager->priv->volume_monitor = g_volume_monitor_get ();
         manager->priv->conf_client = gconf_client_get_default ();
 
         gconf_client_add_dir (manager->priv->conf_client,
@@ -1102,6 +1163,11 @@ gsd_media_keys_manager_stop (GsdMediaKeysManager *manager)
                 priv->conf_client = NULL;
         }
 
+        if (priv->volume_monitor != NULL) {
+                g_object_unref (priv->volume_monitor);
+                priv->volume_monitor = NULL;
+        }
+
         if (priv->connection != NULL) {
                 dbus_g_connection_unref (priv->connection);
                 priv->connection = NULL;
